import { NS } from "@ns"
import { parse } from "./lib/flag";
import { range } from "./util/range";

class Account {
    value: number;
    constructor(value: number) {
        this.value = value;
    }
    /**
     * 
     * @param ns NS
     * @param rate reinvestment rate
     * @param period the time since last update. it's millisecond unit.
     */
    upgradeIncrement(ns: NS, rate: number, period: number): void {
        const production = [...getIndexOfHackNode(ns)]
            .map(x => ns.hacknet.getNodeStats(x).production)
            .reduce((x, y) => x + y);
        this.value += production * rate * (period / 1000);

    }
    formatedValue(ns: NS): string{
        return ns.nFormat(this.value,"$0.0000a");
    }
}

export function getNodeProduction(level: number, ram: number, cores: number, mult: number, bitMult: number): number {
    const moneyGainPerLevel = 1.5;
    const levelMult = level * moneyGainPerLevel;
    const ramMult = Math.pow(1.035, ram - 1);
    const coreMult = ((cores + 5) / 6);
    return levelMult * ramMult * coreMult * mult * bitMult;
}

interface HacknetNodeComponent<T> {
    level: T;
    core: T;
    ram: T;
}

type ComponentKind = "level" | "core" | "ram";

export interface CEInfo {
    upgradeCost: HacknetNodeComponent<number>;
    production: HacknetNodeComponent<number>;
    costEffective: HacknetNodeComponent<number>;
}
interface Factor<T> {
    cost: T;
    production: T;
    costEffective: T;
}

function CEInfoToFactor(ceinfo: CEInfo):HacknetNodeComponent<Factor<number>>{
    return {
        level:{
            cost:ceinfo.upgradeCost.level,
            production:ceinfo.production.level,
            costEffective:ceinfo.costEffective.level,
        },
        core:{
            cost:ceinfo.upgradeCost.core,
            production:ceinfo.production.core,
            costEffective:ceinfo.costEffective.core,
        },
        ram:{
            cost:ceinfo.upgradeCost.ram,
            production:ceinfo.production.ram,
            costEffective:ceinfo.costEffective.ram,
        }
    }
}


function getCostEffective(ns: NS, nodeIndex: number, n: number): CEInfo {
    const coreUpgradeCost = ns.hacknet.getCoreUpgradeCost(nodeIndex, n);
    const ramUpgradeCost = ns.hacknet.getRamUpgradeCost(nodeIndex, n);
    const levelUpgradeCost = ns.hacknet.getLevelUpgradeCost(nodeIndex, n);
    const mult = ns.getHacknetMultipliers().production;
    const bitMult = 1//ns.getBitNodeMultipliers().HacknetNodeMoney;
    const stat = ns.hacknet.getNodeStats(nodeIndex);

    const coreProduction = getNodeProduction(stat.level, stat.ram, stat.cores + n, mult, bitMult) - stat.production;
    const levelProduction = getNodeProduction(stat.level + n, stat.ram, stat.cores, mult, bitMult) - stat.production;
    const ramProduction = getNodeProduction(stat.level, stat.ram * Math.pow(2, n), stat.cores, mult, bitMult) - stat.production;

    return {
        upgradeCost: {
            core: coreUpgradeCost,
            ram: ramUpgradeCost,
            level: levelUpgradeCost,
        },
        production: {
            core: coreProduction,
            ram: ramProduction,
            level: levelProduction,
        },
        costEffective: {
            core: coreProduction / coreUpgradeCost,
            ram: ramProduction / ramUpgradeCost,
            level: levelProduction / levelUpgradeCost
        }
    }
}



function createUpgradeAction(ns: NS, index: number, n: number, kind: ComponentKind): (account: Account) => void {
    return (account: Account) => {
        switch (kind) {
            case "core":
                account.value -= ns.hacknet.getCoreUpgradeCost(index, n);
                ns.hacknet.upgradeCore(index, n)
                break;
            case "level":
                account.value -= ns.hacknet.getLevelUpgradeCost(index, n);
                ns.hacknet.upgradeLevel(index, n)
                break;
            case "ram":
                account.value -= ns.hacknet.getRamUpgradeCost(index, n);
                ns.hacknet.upgradeRam(index, n)
                break;
        }
    }
}

function getMaxCENode(ns: NS, index: number, n: number): [Factor<number>, ComponentKind] {
    const a = getCostEffective(ns, index, n);
    const b = CEInfoToFactor(a);
    const ce = a.costEffective;
    if (ce.core > ce.level) {
        if (ce.core > ce.ram) {
            return [b.core, "core"];
        }
        else {
            return [b.ram, "ram"];
        }
    }
    else {
        if (ce.level > ce.ram) {
            return [b.level, "level"];
        }
        else {
            return [b.ram, "ram"];
        }
    }
}

function* getIndexOfHackNode(ns: NS): Generator<number> {
    const numNodes = ns.hacknet.numNodes()
    for (const it of range(numNodes)) {
        yield it;
    }
}

function getMaxCEIndex(ns: NS, n: number): [number, Factor<number>, ComponentKind] {
    let optimalC: Factor<number> = {
        cost:0,
        production:0,
        costEffective:0,
    };
    let kind: ComponentKind = "core";
    let maxIndex = 0;
    for (const it of getIndexOfHackNode(ns)) {
        const [c, k] = getMaxCENode(ns, it, n);
        if (c.costEffective > optimalC.costEffective) {
            optimalC = c;
            kind = k;
            maxIndex = it;
        }
    }
    return [maxIndex, optimalC, kind];
}

function getBuyNodeCE(ns: NS): Factor<number> {
    const mult = ns.getHacknetMultipliers().production;
    const bitMult = 1//ns.getBitNodeMultipliers().HacknetNodeMoney;
    const cost = ns.hacknet.maxNumNodes() > ns.hacknet.numNodes() ? ns.hacknet.getPurchaseNodeCost() : Infinity;
    const production = getNodeProduction(1, 1, 1, mult, bitMult);
    return {
        cost,
        production,
        costEffective: production/cost,
    };
}

function cycle(ns: NS, account: Account): void {
    const chunkN = 1;

    const [upgradeNodeIndex, upgradeNodeCE, kind] = getMaxCEIndex(ns, chunkN);
    const { costEffective: buyNewCE, cost: buyNewCost } = getBuyNodeCE(ns);
    if (upgradeNodeCE.costEffective > buyNewCE) {
        if(account.value > upgradeNodeCE.cost){
            ns.print(`upgrade ${kind} index ${upgradeNodeIndex} : CE ${upgradeNodeCE.costEffective}`);
            const action = createUpgradeAction(ns, upgradeNodeIndex, chunkN, kind);
            action(account);
            ns.print(`budget ${account.formatedValue(ns)}`);
        }
    }
    else {
        if (account.value > buyNewCost) {
            ns.print(`purchase node : CE ${buyNewCE}`);
            account.value -= buyNewCost;
            ns.hacknet.purchaseNode()
            ns.print(`budget ${account.formatedValue(ns)}`);
        }
    }
}

// eslint-disable-next-line require-await
export async function main(ns: NS): Promise<void> {
    ns.disableLog("ALL")
    const flag = parse(ns.args.map(String));
    if(flag.h || flag.help){
        ns.tprint(`run cmd [-i or --iter] [-r or --rate] [-b --budget]`);
        ns.tprint(`-i : cycle iterating count`);
        ns.tprint(`-r : reinvestment rate`);
        ns.tprint(`-b : initial budget`);
        return;
    }
    const iter = parseInt(flag.i ?? flag.iter ?? "10");
    const rate = parseFloat(flag.r ?? flag.rate ?? "0.5");
    const budget = parseInt(flag.b ?? flag.budget ?? "0");

    const account = new Account(budget);
    ns.tail();
    const period = 1000;
    for (let i = 0; i < iter; i++) {
        //ns.clearLog()
        cycle(ns, account);
        await ns.sleep(period);
        account.upgradeIncrement(ns,rate,period);
    }
}