import { NS, Server } from "@ns"
import { calculateServerGrowth, FormulaHelper } from "./formula";


export type BatchType = "hack" | "grow" | "weaken" | "weakenLoop";

type BatchFilePackEntry = {
    type: BatchType;
    path: string;
    content: string;
    ram: number;
};

type FilePack = BatchFilePackEntry[];

const createScriptContent: (f: string) => string = (funcName: string) => `export async function main(ns) {
    if(ns.args.length < 3){
        ns.print("[ERROR] args required");
        ns.print("run cmd [sleepDuration] [hostname] [effectStock]")
        ns.exit();
        return;
    }
    const sleepDuration = parseInt(ns.args[0]);
    const hostname = ns.args[1].toString();
    const stock = ns.args[2] == "true";
    if(sleepDuration > 0){
        await ns.sleep(sleepDuration);
    }
    await ns.${funcName}(hostname,{
        stock,
    });
}`

async function unpackFileEntry(ns: NS, filePack: BatchFilePackEntry): Promise<void> {
    await ns.write(filePack.path, filePack.content, "w");
}

export async function unpackFilePack(ns: NS, pack: FilePack): Promise<void> {
    for (const entry of pack) {
        await unpackFileEntry(ns, entry);
    }
}

export const batchFilePack: FilePack = [
    {
        type: "hack",
        path: "/tmp/hack.js",
        content: createScriptContent("hack"),
        ram:1.7,
    },
    {
        type: "weaken",
        path: "/tmp/weaken.js",
        content: createScriptContent("weaken"),
        ram: 1.75,
    },
    {
        type: "grow",
        path: "/tmp/grow.js",
        content: createScriptContent("grow"),
        ram: 1.75,
    },
    {
        type:"weakenLoop",
        path: "/tmp/weakenLoop.js",
        ram: 1.75,
        content: `export async function main(ns) {
    if(ns.args.length < 3){
        ns.print("[ERROR] args required");
        ns.print("run cmd [sleepDuration] [hostname] [effectStock]")
        ns.exit();
        return;
    }
    const sleepDuration = parseInt(ns.args[0]);
    const hostname = ns.args[1].toString();
    const stock = ns.args[2] == "true";
    for(;;){
        await ns.weaken(hostname,{
            stock,
        });
    }
}`,
    }
]

export async function scpBatchFilePack(ns: NS, hostname: string): Promise<void> {
    const list = batchFilePack.map(entry => entry.path);
    await installBatchFilePack(ns);
    await ns.scp(list, hostname);
}

export async function installBatchFilePack(ns: NS): Promise<void> {
    await unpackFilePack(ns, batchFilePack);
}

export function isBatchFilePackInstalled(ns: NS, hostname?: string): boolean {
    return batchFilePack.every(entry => ns.fileExists(entry.path, hostname))
}

export interface ExecOption {
    hostname: string;
    /**
     * @default 1
     */
    numThread?: number;
    /**
     * @default 0
     */
    sleepDuration?: number;
    /**
     * target hostname to operate "hack", "weaken" or "grow"
     */
    target: string;
    stock?: boolean;
}

function execBatchfile(ns: NS, path: string, option: ExecOption): number {
    return ns.exec(path, option.hostname, option.numThread,
        option.sleepDuration ?? 0, option.target, option.stock ?? false);
}

export function execHack(ns: NS, option: ExecOption): number {
    return execBatchfile(ns, "/tmp/hack.js", option);
}
export function execGrow(ns: NS, option: ExecOption): number {
    return execBatchfile(ns, "/tmp/grow.js", option);
}
export function execWeaken(ns: NS, option: ExecOption): number {
    return execBatchfile(ns, "/tmp/weaken.js", option);
}
export function execWeakenLoop(ns: NS, option: ExecOption): number {
    return execBatchfile(ns, "/tmp/weakenLoop.js", option);
}
export function getRamOfScript(ns: NS, type: BatchType): number{
    return batchFilePack.filter(x=>x.type == type)[0].ram;
}

//type BatchType = "hack" | "grow" | "weaken";
//
//type Batch = {
//    type: BatchType;
//    options: ExecOption;
//}
type ReserveRequest = {
    hostname: string;
    usableRam: number;
}

function getKeyOfHWGWBatchManager(hostname: string): string{
    return `HWGW-${hostname}`;
}
interface HWGW {
    HThreadCount: number;
    HWThreadCount: number;
    GThreadCount: number;
    GWThreadCount: number;
}

type WorkReserved = {
    allocated:false;
} | {
    allocated:true;
    hwgw: HWGW;
};

type Resource = {
    hostname: string;
    usableRam: string;
};

class HWGWBatchManager {
    willExecuted: WorkReserved[];
    server: Server;
    ns: NS;
    createdAt: number;


    constructor(ns: NS, hostname: string) {
        this.ns = ns;
        this.willExecuted = [];
        this.server = ns.getServer(hostname);
        this.createdAt = ns.getTimeSinceLastAug();
        
        this.growArray();
    }
    private popRequest(): ReserveRequest[]{
        const key = getKeyOfHWGWBatchManager(this.server.hostname);
        const text = localStorage.getItem(key)
        if(text === null){
            return [];
        }
        localStorage.setItem(key,"");
        return JSON.parse(text) as ReserveRequest[];
    }
    private growArray(n: number): void {
        while (this.willExecuted.length <= n) {
            this.willExecuted.push({
                allocated:false
            });
        }
    }

    reserveBatch(time: number): void {
        const milli = time - this.ns.getTimeSinceLastAug();
        const sec = Math.floor(milli / 1000);
        this.growArray(sec);
        this.willExecuted[sec] = true;
    }
    async tick(): Promise<void> {
        const data = this.willExecuted.shift();
        await this.sleepCorrect();
    }
    /**
     * exact time Sleep
     */
    async sleepCorrect(): Promise<void> {
        const cur = this.ns.getTimeSinceLastAug();
        const sliceTime = 250;
        const delta = (cur - this.createdAt + sliceTime/2) % sliceTime - sliceTime/2;
        await this.ns.sleep(sliceTime - delta);
    }
}


interface HackCalculation extends HWGW {
    earnMoneyPerSec: number;
    earnExpPerSec: number;
    /**
     * cycle time in seconds.
     */
    cycleTime: number;
    totalThreadCount: number;
}

export function calculateBatchResource(ns: NS, hostname: string, hackThread: number, startMoney?: number, cores = 1): HackCalculation {
    const helper = new FormulaHelper(ns);
    const server = ns.getServer(hostname);
    startMoney ??= server.moneyMax;
    server.hackDifficulty = server.minDifficulty;

    const percentMoneyHacked = helper.calculatePercentMoneyHacked(server);
    const hackingChance = helper.calculateHackingChance(server);
    const earnMoneyPerCycle = hackThread * percentMoneyHacked * startMoney * hackingChance;
    const earnExpPerCycle = helper.calculateHackingExpGain(server) * hackThread * hackingChance;
    const cycleTime = 1 + helper.calculateWeakenTime(server);

    const increasedSecurityAfterHack = helper.calculateHackSecurity(hackThread) + server.minDifficulty;
    const nWeakenHackThreads = Math.ceil(helper.calculateWeakenThreadCountToTargetSecurity(
        server.minDifficulty,
        increasedSecurityAfterHack, cores));

    const nGrowThreads = Math.ceil(helper.calculateGrowThreadCount(hostname, startMoney - earnMoneyPerCycle, startMoney, cores));
    const increasedSecurityAfterGrow = helper.calculuateGrowthSecurity(nGrowThreads, cores) + server.minDifficulty;

    const nWeakenGrowThreads = Math.ceil(helper.calculateWeakenThreadCountToTargetSecurity(
        server.minDifficulty,
        increasedSecurityAfterGrow, cores));

    return {
        earnMoneyPerSec: earnMoneyPerCycle / cycleTime,
        earnExpPerSec: earnExpPerCycle / cycleTime,
        cycleTime,
        totalThreadCount: hackThread + nWeakenHackThreads + nGrowThreads + nWeakenGrowThreads,
        HThreadCount: hackThread,
        HWThreadCount: nWeakenHackThreads,
        GThreadCount: nGrowThreads,
        GWThreadCount: nWeakenGrowThreads
    }
}

function calculateOptimal(ns: NS, target: string, req: ReserveRequest): HWGW {

    return {
        GThreadCount: 0,
        HThreadCount: 0,
        GWThreadCount: 0,
        HWThreadCount: 0,
    }
}

export function requestHWGW(req: ReserveRequest, target: string): void {
    const key = getKeyOfHWGWBatchManager(target);
    const data = localStorage.getItem(key);
    let requests: ReserveRequest[];
    if (data) {
        requests = JSON.parse(data);
    }
    else {
        requests = [];
    }
    requests.push(req);
    localStorage.setItem(key, JSON.stringify(requests));
}

//export async function batchDaemon(ns: NS): Promise<void> {
//    const mgr = new HWGWBatchManager(ns, BATCH_KEY);
//    for (; ;) {
//        await mgr.tick();
//    }
//}