import { NS } from "@ns";

import { subarrayMaxSolve } from "./solve/submaxarr";
import { largestPrimeSolve } from "./solve/largestPrime";
import { spiralSolve } from "./solve/spiral";
import { arrayJumpSolve } from "./solve/arrayJump";
import { arrayJump2Solve } from "./solve/arrayJump2";
import { minimumTrianglePathSumSolve} from "./solve/minsumTriangle"
import { graph2coloringSolve } from "./solve/graph2color";
import { shortestPathInGridSolve } from "./solve/shortestPathInGrid";
import { algorithmicTradeIIISolve, algorithmicTradeIISolve, algorithmicTradeISolve, algorithmicTradeIVSolve } from "./solve/algorithmicTrader";

import {selectAllServerList} from "lib/servers";
import { logFile } from "./log";
import { SolveFailError } from "./solve/unsolved";
import { RLECompressionSolve, decompressLZSolve, compressLZSolve } from "./solve/compression";

export { SolveFailError } from "./solve/unsolved";

export interface ContractSolver{
    name: string;
    solve(n :any): string[] | number;
}

// eslint-disable-next-line @typescript-eslint/no-unused-vars
const ContractMAP: {[key:string]:ContractSolver} = {
    /**
     * A prime factor is a factor that is a prime number.
     * What is the largest prime factor of the given number?
     */
    "Find Largest Prime Factor": {
        name: "Find Largest Prime Factor",
        solve: largestPrimeSolve
    },
    /**
     *  Given the following integer array, find the contiguous subarray
     *  (containing at least one number) which has the largest sum and return that sum.
     *  'Sum' refers to the sum of all the numbers in the subarray.
     */
    "Subarray with Maximum Sum": {
        name: "Subarray with Maximum Sum",
        solve: subarrayMaxSolve,
    },
    /*"Total Ways to Sum": {
        name: "Total Ways to Sum",
        solve() {
            notImplemented();
        }
    },
    "Total Ways to Sum II": {
        name: "Total Ways to Sum II",
        solve() {
            notImplemented();
        }
    },*/
    "Spiralize Matrix": {
        name: "Spiralize Matrix",
        solve: spiralSolve,
    },
    "Array Jumping Game": {
        name: "Array Jumping Game",
        solve: arrayJumpSolve,
    },
    "Array Jumping Game II": {
        name: "Array Jumping Game II",
        solve: arrayJump2Solve
    },/*
    "Merge Overlapping Intervals": {
        name: "Merge Overlapping Intervals",
        solve() {
            notImplemented();
        }
    },
    "Generate IP Addresses": {
        name: "Generate IP Addresses",
        solve() {
            notImplemented();
        }
    },*/
    "Algorithmic Stock Trader I": {
        name: "Algorithmic Stock Trader I",
        solve: algorithmicTradeISolve,
    },
    "Algorithmic Stock Trader II": {
        name: "Algorithmic Stock Trader II",
        solve: algorithmicTradeIISolve,
    },
    "Algorithmic Stock Trader III": {
        name: "Algorithmic Stock Trader III",
        solve: algorithmicTradeIIISolve
    },
    "Algorithmic Stock Trader IV": {
        name: "Algorithmic Stock Trader IV",
        solve: algorithmicTradeIVSolve
    },
    "Minimum Path Sum in a Triangle": {
        name: "Minimum Path Sum in a Triangle",
        solve: minimumTrianglePathSumSolve,
    },/*
    "Unique Paths in a Grid I": {
        name: "Unique Paths in a Grid I",
        solve() {
            notImplemented();
        }
    },
    "Unique Paths in a Grid II": {
        name: "Unique Paths in a Grid II",
        solve() {
            notImplemented();
        }
    },*/
    "Shortest Path in a Grid": {
        name: "Shortest Path in a Grid",
        solve: shortestPathInGridSolve
    },/*
    "Sanitize Parentheses in Expression": {
        name: "Sanitize Parentheses in Expression",
        solve() {
            notImplemented();
        }
    },
    "Find All Valid Math Expressions": {
        name: "Find All Valid Math Expressions",
        solve() {
            notImplemented();
        }
    },
    "HammingCodes: Integer to Encoded Binary": {
        name: "HammingCodes: Integer to Encoded Binary",
        solve() {
            notImplemented();
        }
    },
    "HammingCodes: Encoded Binary to Integer": {
        name: "HammingCodes: Encoded Binary to Integer",
        solve() {
            notImplemented();
        }
    },*/
    "Proper 2-Coloring of a Graph": {
        name: "Proper 2-Coloring of a Graph",
        solve: graph2coloringSolve
    },
    "Compression I: RLE Compression": {
        name: "Compression I: RLE Compression",
        solve: RLECompressionSolve,
    },
    "Compression II: LZ Decompression": {
        name: "Compression II: LZ Decompression",
        solve: decompressLZSolve,
    },
    "Compression III: LZ Compression": {
        name: "Compression III: LZ Compression",
        solve: compressLZSolve,
    },/*
    "Encryption I: Caesar Cipher": {
        name: "Encryption I: Caesar Cipher",
        solve() {
            notImplemented();
        }
    },
    "Encryption II: Vigenère Cipher": {
        name: "Encryption II: Vigenère Cipher",
        solve() {
            notImplemented();
        }
    }*/
}

export function isSolvable(ns:NS,filename: string, hostname: string): boolean {
    const ct = ns.codingcontract;
    const type = ct.getContractType(filename,hostname);
    return (type in ContractMAP)
}

export async function solveContract(ns:NS,filename: string, hostname: string): Promise<[boolean,string]>{
    const ct = ns.codingcontract;
    const type = ct.getContractType(filename,hostname);
    if(type in ContractMAP){
        const solver = ContractMAP[type];
        const data = ct.getData(filename,hostname);
        let ans: number|string[];
        try {
            ans = solver.solve(data);
        } catch (error) {
            if(error instanceof SolveFailError){
                ns.print(`[ERROR] failed to solve problem ${type}. throw SolveFailError`);
                ns.toast(`failed to solve problem ${type}`,"error");
                await logFile(ns,type+"-Fail",data);
                return [false,""];
            }
            else{
                throw error;
            }
        }
        const reward = ct.attempt(ans,filename,hostname,{returnReward:true});
        if(reward === ""){
            ns.print(`[ERROR] failed to solve problem ${type}`);
            ns.toast(`failed to solve problem ${type}`,"error");
            await logFile(ns,type,[data,ans]);
            return [false,""];
        }
        else {
            ns.print(`success to solve.\n${reward}`);
        }
        return [true,reward.toString()];
    }
    return [false,""];
}

export interface RemoteFilePath{
    hostname: string;
    filename: string;
}

export function getContractList(ns: NS, hostname: string): RemoteFilePath[]{
    const ctFilenameList = ns.ls(hostname,".cct").filter(x=>x.endsWith(".cct"));
    return ctFilenameList.map(filename=>({
            hostname: hostname,
            filename: filename
        }));
}

export function selectAllContract(ns: NS): RemoteFilePath[] {
    const serverList = selectAllServerList(ns);
    const ctList = serverList.map(server=>getContractList(ns,server.hostname))
                        .flat();
    return ctList;
}